/*******************************************************************************
*
* DESCRIPTION:  Optimized C implementation of Crypton version 1.0
*
* AUTHOR:       Chae Hoon Lim and Hyo Sun Hwang
*               E-mail: chlim@future.co.kr, hyosun@future.co.kr
*               http://crypt.future.co.kr/~chlim
*               Information and Communications Research Center,
*               Future Systems, Inc.
*               372-2, Yang Jae-Dong, Seo Cho-Gu, Seoul, 137-130, KOREA
*               Tel: +82-2-578-0581 (ext. 557), Fax: +82-2-578-0584
*
* Some recoding by Alexander Myasnikow
* Web: www.darksoftware.narod.ru
*
*******************************************************************************/

#define USE_ASM

#include "crypton.h"
#include "crypton.tab"

/*============= Definitions for Encryption/Decryption =============*/

/* byte extracting/positioning from/into DWORD */

#define GetB0(A)  ( (BYTE)((A)    ) )   /* A: DWORD */
#define GetB1(A)  ( (BYTE)((A)>> 8) )
#define GetB2(A)  ( (BYTE)((A)>>16) )
#define GetB3(A)  ( (BYTE)((A)>>24) )

#define PutB0(B)  ( (DWORD)(B)     )    /* B: BYTE */
#define PutB1(B)  ( (DWORD)(B)<< 8 )
#define PutB2(B)  ( (DWORD)(B)<<16 )
#define PutB3(B)  ( (DWORD)(B)<<24 )


/* Round transformation */
#define CryptonRound(A,B,C,D, E,F,G,H, K) {       \
    A  = SS0[GetB0(E)] ^ SS1[GetB0(F)];           \
    B  = SS1[GetB1(E)] ^ SS2[GetB1(F)];           \
    C  = SS2[GetB2(E)] ^ SS3[GetB2(F)];           \
    D  = SS3[GetB3(E)] ^ SS0[GetB3(F)];           \
    A ^= SS2[GetB0(G)] ^ SS3[GetB0(H)] ^ (K)[0];  \
    B ^= SS3[GetB1(G)] ^ SS0[GetB1(H)] ^ (K)[1];  \
    C ^= SS0[GetB2(G)] ^ SS1[GetB2(H)] ^ (K)[2];  \
    D ^= SS1[GetB3(G)] ^ SS2[GetB3(H)] ^ (K)[3];  \
}


/* special final round */
#define CryptonFinal(pb, A,B,C,D, K) {                \
    pb[ 0] = (BYTE)(S2[GetB0(A)] ^ GetB0((K)[0]));    \
    pb[ 1] = (BYTE)(S3[GetB0(B)] ^ GetB1((K)[0]));    \
    pb[ 4] = (BYTE)(S3[GetB1(A)] ^ GetB0((K)[1]));    \
    pb[ 5] = (BYTE)(S0[GetB1(B)] ^ GetB1((K)[1]));    \
    pb[ 8] = (BYTE)(S0[GetB2(A)] ^ GetB0((K)[2]));    \
    pb[ 9] = (BYTE)(S1[GetB2(B)] ^ GetB1((K)[2]));    \
    pb[12] = (BYTE)(S1[GetB3(A)] ^ GetB0((K)[3]));    \
    pb[13] = (BYTE)(S2[GetB3(B)] ^ GetB1((K)[3]));    \
    pb[ 2] = (BYTE)(S0[GetB0(C)] ^ GetB2((K)[0]));    \
    pb[ 3] = (BYTE)(S1[GetB0(D)] ^ GetB3((K)[0]));    \
    pb[ 6] = (BYTE)(S1[GetB1(C)] ^ GetB2((K)[1]));    \
    pb[ 7] = (BYTE)(S2[GetB1(D)] ^ GetB3((K)[1]));    \
    pb[10] = (BYTE)(S2[GetB2(C)] ^ GetB2((K)[2]));    \
    pb[11] = (BYTE)(S3[GetB2(D)] ^ GetB3((K)[2]));    \
    pb[14] = (BYTE)(S3[GetB3(C)] ^ GetB2((K)[3]));    \
    pb[15] = (BYTE)(S0[GetB3(D)] ^ GetB3((K)[3]));    \
}

/* save inter-round values */
#if defined(CheckInterValue)
extern BYTE pbInterVal[NO_ROUNDS + 1][BLOCK_SIZE];
extern int PRN;
#define INTER_VALUE_OUT(A, B, C, D, N) {    \
        if(PRN) {                               \
            DWORD temp[4];                      \
            temp[0] = A;                        \
            temp[1] = B;                        \
            temp[2] = C;                        \
            temp[3] = D;                        \
            memcpy(pbInterVal[N], (BYTE *)temp, BLOCK_SIZE); \
        }                                       \
    }
#else
#define INTER_VALUE_OUT(A, B, C, D, N)
#endif


/*######################## Block Encrypt/Decrypt #########################*/


/*

void CryptonEncrypt(BYTE *pbData, DWORD *pdwRoundKey)
{
    DWORD A, B, C, D, E, F, G, H, *K=pdwRoundKey;

    E = ((DWORD *)pbData)[0] ^ K[0];
    F = ((DWORD *)pbData)[1] ^ K[1];
    G = ((DWORD *)pbData)[2] ^ K[2];
    H = ((DWORD *)pbData)[3] ^ K[3];
    INTER_VALUE_OUT(E,F,G,H, 0);

    CryptonRound(A,B,C,D, E,F,G,H, K+ 4);
    INTER_VALUE_OUT(A,B,C,D, 1);
    CryptonRound(E,F,G,H, C,D,A,B, K+ 8);
    INTER_VALUE_OUT(E,F,G,H, 2);
    CryptonRound(A,B,C,D, E,F,G,H, K+12);
    INTER_VALUE_OUT(A,B,C,D, 3);
    CryptonRound(E,F,G,H, C,D,A,B, K+16);
    INTER_VALUE_OUT(E,F,G,H, 4);
    CryptonRound(A,B,C,D, E,F,G,H, K+20);
    INTER_VALUE_OUT(A,B,C,D, 5);
    CryptonRound(E,F,G,H, C,D,A,B, K+24);
    INTER_VALUE_OUT(E,F,G,H, 6);
    CryptonRound(A,B,C,D, E,F,G,H, K+28);
    INTER_VALUE_OUT(A,B,C,D, 7);
    CryptonRound(E,F,G,H, C,D,A,B, K+32);
    INTER_VALUE_OUT(E,F,G,H, 8);
    CryptonRound(A,B,C,D, E,F,G,H, K+36);
    INTER_VALUE_OUT(A,B,C,D, 9);
    CryptonRound(E,F,G,H, C,D,A,B, K+40);
    INTER_VALUE_OUT(E,F,G,H,10);
    CryptonRound(A,B,C,D, E,F,G,H, K+44);
    INTER_VALUE_OUT(A,B,C,D,11);
    CryptonFinal(pbData, A,B,C,D, K+48);
    INTER_VALUE_OUT(((DWORD *)pbData)[0], ((DWORD *)pbData)[1], \
                    ((DWORD *)pbData)[2], ((DWORD *)pbData)[3], 12);
}

*/


/**************** MACROS for ENCRYPTION ***********************************/
#define     TEMP_MEM_0      dword ptr [esp+0x00]
#define     TEMP_MEM_1      dword ptr [esp+0x04]
#define     TEMP_MEM_2      dword ptr [esp+0x08]
#define     TEMP_MEM_3      dword ptr [esp+0x0C]
#define     KEY_ADDRESS     dword ptr [esp+0x10]
#define     DATA_ADDRESS    dword ptr [esp+0x14]
#define     TEMP_EBP        dword ptr [esp+0x18]

/****************************************/
#define ASM_CryptonRound(K0,K1,K2,K3,S0,S1,S2,S3) {         \
    _asm    mov     TEMP_MEM_1, esi             /*  0S  */  \
    _asm    mov     bl, al                      /*  1A  */  \
    _asm    mov     dl, ah                      /*  1B  */  \
    _asm    mov     TEMP_MEM_2, ecx             /*  0S  */  \
    _asm    shr     eax, 0x10                   /*  1   */  \
    _asm    mov     ebp, dword ptr [S0+4*ebx]   /*  1A  */  \
\
    _asm    mov     TEMP_MEM_3, edi             /*  0S  */  \
    _asm    mov     bl, al                      /*  1C  */  \
    _asm    mov     esi, dword ptr [S1+4*edx]   /*  1B  */  \
    _asm    mov     dl, ah                      /*  1D  */  \
    _asm    mov     eax, TEMP_MEM_1             /*  2   */  \
    _asm    mov     ecx, dword ptr [S2+4*ebx]   /*  1C  */  \
\
    _asm    mov     edi, dword ptr [S3+4*edx]   /*  1D  */  \
    _asm    mov     bl, al                      /*  2A  */  \
    _asm    mov     dl, ah                      /*  2B  */  \
    _asm    shr     eax, 0x10                   /*  2   */  \
    _asm    xor     ebp, dword ptr [S1+4*ebx]   /*  2A  */  \
    _asm    mov     bl, al                      /*  2C  */  \
\
    _asm    xor     esi, dword ptr [S2+4*edx]   /*  2B  */  \
    _asm    mov     dl, ah                      /*  2D  */  \
    _asm    mov     eax, TEMP_MEM_2             /*  3   */  \
    _asm    xor     ecx, dword ptr [S3+4*ebx]   /*  2C  */  \
    _asm    mov     bl, al                      /*  3A  */  \
    _asm    xor     edi, dword ptr [S0+4*edx]   /*  2D  */  \
\
    _asm    mov     dl, ah                      /*  3B  */  \
    _asm    shr     eax, 0x10                   /*  3   */  \
    _asm    xor     ebp, dword ptr [S2+4*ebx]   /*  3A  */  \
    _asm    mov     bl, al                      /*  3C  */  \
    _asm    xor     esi, dword ptr [S3+4*edx]   /*  3B  */  \
    _asm    mov     dl, ah                      /*  3D  */  \
\
    _asm    mov     eax, TEMP_MEM_3             /*  4   */  \
    _asm    xor     ecx, dword ptr [S0+4*ebx]   /*  3C  */  \
    _asm    mov     bl, al                      /*  4A  */  \
    _asm    xor     edi, dword ptr [S1+4*edx]   /*  3D  */  \
    _asm    mov     dl, ah                      /*  4B  */  \
    _asm    shr     eax, 0x10                   /*  4   */  \
\
    _asm    xor     ebp, dword ptr [S3+4*ebx]   /*  4A  */  \
    _asm    mov     bl, al                      /*  4D  */  \
    _asm    xor     esi, dword ptr [S0+4*edx]   /*  4B  */  \
    _asm    mov     dl, ah                      /*  4C  */  \
    _asm    mov     eax, ebp                    /*  K.  */  \
    _asm    mov     ebp, KEY_ADDRESS            /*  K.  */  \
\
    _asm    xor     ecx, dword ptr [S1+4*ebx]   /*  4D  */  \
    _asm    xor     edi, dword ptr [S2+4*edx]   /*  4C  */  \
    _asm    xor     eax, dword ptr [ebp+K0]     /*  K0  */  \
    _asm    xor     esi, dword ptr [ebp+K1]     /*  K1  */  \
    _asm    xor     ecx, dword ptr [ebp+K2]     /*  K2  */  \
    _asm    xor     edi, dword ptr [ebp+K3]     /*  k3  */  \
}

/****************************************/
#define ASM_Init(K0,K1,K2,K3) {                             \
    _asm    sub     esp, 0x1C                   /*1     */  \
    _asm    mov     ebx, dword ptr [pbData]     /*2     */  \
    _asm    mov     edx, dword ptr [pdwRoundKey]/*3     */  \
\
    _asm    mov     TEMP_EBP, ebp               /*4     */  \
    _asm    mov     eax, dword ptr [ebx+0x00]   /*5     */  \
    _asm    mov     esi, dword ptr [ebx+0x04]   /*6     */  \
    _asm    mov     DATA_ADDRESS, ebx           /*1     */  \
    _asm    mov     ecx, dword ptr [ebx+0x08]   /*2     */  \
    _asm    mov     edi, dword ptr [ebx+0x0C]   /*3     */  \
    _asm    mov     KEY_ADDRESS, edx            /*4     */  \
    _asm    mov     ebx, dword ptr [edx+0x00]   /*5     */  \
    _asm    mov     ebp, dword ptr [edx+0x04]   /*6     */  \
\
    _asm    xor     ecx, dword ptr [edx+0x08]   /*1     */  \
    _asm    xor     eax, ebx                    /*2     */  \
    _asm    xor     esi, ebp                    /*3     */  \
    _asm    xor     edi, dword ptr [edx+0x0C]   /*4     */  \
    _asm    xor     ebx, ebx                    /*5     */  \
    _asm    xor     edx, edx                    /*6     */  \
}

/****************************************/
#define ASM_Final(K0,K1,K2,K3,S0,S1,S2,S3) {                \
    _asm    mov     dword ptr [esp+0x08], ecx   /*      */  \
    _asm    xor     ebx, ebx                    /*      */  \
    _asm    xor     edx, edx                    /*      */  \
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     dl, ah                      /*      */  \
    _asm    mov     ebp, DATA_ADDRESS           /*      */  \
\
    _asm    shr     eax, 0x10                   /*      */  \
    _asm    mov     cl, byte ptr [S2+ebx]       /*      */  \
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     ch, byte ptr [S3+edx]       /*      */  \
    _asm    mov     dl, ah                      /*      */  \
    _asm    mov     byte ptr [ebp+0x00], cl     /*      */  \
\
    _asm    mov     eax, esi                    /*      */  \
    _asm    mov     byte ptr [ebp+0x04], ch     /*      */  \
    _asm    mov     cl, byte ptr [S0+ebx]       /*      */  \
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     ch, byte ptr [S1+edx]       /*      */  \
    _asm    mov     dl, ah                      /*      */  \
\
    _asm    shr     eax, 0x10                   /*      */  \
    _asm    mov     byte ptr [ebp+0x08], cl     /*      */  \
    _asm    mov     byte ptr [ebp+0x0C], ch     /*      */  \
    _asm    mov     cl, byte ptr [S3+ebx]       /*      */  \
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     ch, byte ptr [S0+edx]       /*      */  \
\
    _asm    mov     dl, ah                      /*      */  \
    _asm    mov     byte ptr [ebp+0x01], cl     /*      */  \
    _asm    mov     eax, dword ptr [esp+0x08]   /*      */  \
    _asm    mov     byte ptr [ebp+0x05], ch     /*      */  \
    _asm    mov     cl, byte ptr [S1+ebx]       /*      */  \
    _asm    mov     bl, al                      /*      */  \
\
    _asm    mov     ch, byte ptr [S2+edx]       /*      */  \
    _asm    mov     dl, ah                      /*      */  \
    _asm    mov     byte ptr [ebp+0x09], cl     /*      */  \
    _asm    mov     byte ptr [ebp+0x0D], ch     /*      */  \
    _asm    shr     eax, 0x10                   /*      */  \
    _asm    mov     cl, byte ptr [S0+ebx]       /*      */  \
\
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     ch, byte ptr [S1+edx]       /*      */  \
    _asm    mov     dl, ah                      /*      */  \
    _asm    mov     byte ptr [ebp+0x02], cl     /*      */  \
    _asm    mov     eax, edi                    /*      */  \
    _asm    mov     byte ptr [ebp+0x06], ch     /*      */  \
\
    _asm    mov     cl, byte ptr [S2+ebx]       /*      */  \
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     ch, byte ptr [S3+edx]       /*      */  \
    _asm    mov     dl, ah                      /*      */  \
    _asm    shr     eax, 0x10                   /*      */  \
    _asm    mov     byte ptr [ebp+0x0A], cl     /*      */  \
\
    _asm    mov     byte ptr [ebp+0x0E], ch     /*      */  \
    _asm    mov     cl, byte ptr [S1+ebx]       /*      */  \
    _asm    mov     bl, al                      /*      */  \
    _asm    mov     ch, byte ptr [S2+edx]       /*      */  \
    _asm    mov     dl, ah                      /*      */  \
    _asm    mov     byte ptr [ebp+0x03], cl     /*      */  \
\
    _asm    mov     byte ptr [ebp+0x07], ch     /*      */  \
    _asm    mov     cl, byte ptr [S3+ebx]       /*      */  \
    _asm    mov     esi, KEY_ADDRESS            /*      */  \
    _asm    mov     ch, byte ptr [S0+edx]       /*      */  \
    _asm    mov     byte ptr [ebp+0x0B], cl     /*      */  \
    _asm    mov     byte ptr [ebp+0x0F], ch     /*      */  \
\
    _asm    mov     eax, dword ptr [ebp+0x00]   /*      */  \
    _asm    mov     ebx, dword ptr [ebp+0x04]   /*      */  \
    _asm    mov     ecx, dword ptr [esi+K2]     /*      */  \
    _asm    mov     edi, dword ptr [esi+K3]     /*      */  \
    _asm    xor     eax, dword ptr [esi+K0]     /*      */  \
    _asm    xor     ebx, dword ptr [esi+K1]     /*      */  \
\
    _asm    xor     ecx, dword ptr [ebp+0x08]   /*      */  \
    _asm    xor     edi, dword ptr [ebp+0x0C]   /*      */  \
    _asm    mov     dword ptr [ebp+0x00], eax   /*      */  \
    _asm    mov     dword ptr [ebp+0x04], ebx   /*      */  \
    _asm    mov     dword ptr [ebp+0x08], ecx   /*      */  \
    _asm    mov     dword ptr [ebp+0x0C], edi   /*      */  \
\
    _asm    mov     ebp, TEMP_EBP               /*3     */  \
    _asm    add     esp, 0x1C                   /*2     */  \
}

/***********************************************************************/
/***    Encrypt a 128-bit block P and save the result back to P      ***/
/***********************************************************************/

DWORD EncKey[52];
DWORD DecKey[52];

DWORD *pdwRoundKey;

void __stdcall __export
crypt (BYTE * pbData)
{

  pdwRoundKey = &EncKey;

  ASM_Init (0x00, 0x04, 0x08, 0x0C);
  ASM_CryptonRound (0x10, 0x14, 0x18, 0x1C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x20, 0x24, 0x28, 0x2C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x30, 0x34, 0x38, 0x3C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x40, 0x44, 0x48, 0x4C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x50, 0x54, 0x58, 0x5C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x60, 0x64, 0x68, 0x6C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x70, 0x74, 0x78, 0x7C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x80, 0x84, 0x88, 0x8C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x90, 0x94, 0x98, 0x9C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0xA0, 0xA4, 0xA8, 0xAC, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0xB0, 0xB4, 0xB8, 0xBC, SS0, SS1, SS2, SS3);
  ASM_Final (0xC0, 0xC4, 0xC8, 0xCC, S0, S1, S2, S3);
}

void __stdcall __export
decrypt (BYTE * pbData)
{

  pdwRoundKey = &DecKey;

  ASM_Init (0x00, 0x04, 0x08, 0x0C);
  ASM_CryptonRound (0x10, 0x14, 0x18, 0x1C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x20, 0x24, 0x28, 0x2C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x30, 0x34, 0x38, 0x3C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x40, 0x44, 0x48, 0x4C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x50, 0x54, 0x58, 0x5C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x60, 0x64, 0x68, 0x6C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x70, 0x74, 0x78, 0x7C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0x80, 0x84, 0x88, 0x8C, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0x90, 0x94, 0x98, 0x9C, SS0, SS1, SS2, SS3);
  ASM_CryptonRound (0xA0, 0xA4, 0xA8, 0xAC, SS2, SS3, SS0, SS1);
  ASM_CryptonRound (0xB0, 0xB4, 0xB8, 0xBC, SS0, SS1, SS2, SS3);
  ASM_Final (0xC0, 0xC4, 0xC8, 0xCC, S0, S1, S2, S3);
}



DWORD pdwExpKey[8];


/*######################### Expand a User Key ########################*/

/* Generate 8 Expanded keys from a user-supplied key */

void __stdcall __export
setup (BYTE * pbUserKey)
{
  DWORD UK[8], Sum0, Sum1;

  BlockCopy (UK, pbUserKey);
  BlockCopy (UK + 4, pbUserKey + 16);

  /* odd bytes */
  pdwExpKey[0] = SS3[GetB0 (UK[6])] ^ SS2[GetB0 (UK[4])] ^
    SS1[GetB0 (UK[2])] ^ SS0[GetB0 (UK[0])];
  pdwExpKey[1] = SS0[GetB2 (UK[6])] ^ SS3[GetB2 (UK[4])] ^
    SS2[GetB2 (UK[2])] ^ SS1[GetB2 (UK[0])];
  pdwExpKey[2] = SS1[GetB0 (UK[7])] ^ SS0[GetB0 (UK[5])] ^
    SS3[GetB0 (UK[3])] ^ SS2[GetB0 (UK[1])];
  pdwExpKey[3] = SS2[GetB2 (UK[7])] ^ SS1[GetB2 (UK[5])] ^
    SS0[GetB2 (UK[3])] ^ SS3[GetB2 (UK[1])];
  Sum0 = pdwExpKey[0] ^ pdwExpKey[1] ^ pdwExpKey[2] ^ pdwExpKey[3];

  /* even bytes */
  pdwExpKey[4] = SS1[GetB1 (UK[6])] ^ SS0[GetB1 (UK[4])] ^
    SS3[GetB1 (UK[2])] ^ SS2[GetB1 (UK[0])];
  pdwExpKey[5] = SS2[GetB3 (UK[6])] ^ SS1[GetB3 (UK[4])] ^
    SS0[GetB3 (UK[2])] ^ SS3[GetB3 (UK[0])];
  pdwExpKey[6] = SS3[GetB1 (UK[7])] ^ SS2[GetB1 (UK[5])] ^
    SS1[GetB1 (UK[3])] ^ SS0[GetB1 (UK[1])];
  pdwExpKey[7] = SS0[GetB3 (UK[7])] ^ SS3[GetB3 (UK[5])] ^
    SS2[GetB3 (UK[3])] ^ SS1[GetB3 (UK[1])];
  Sum1 = pdwExpKey[4] ^ pdwExpKey[5] ^ pdwExpKey[6] ^ pdwExpKey[7];

  pdwExpKey[0] ^= Sum1;
  pdwExpKey[1] ^= Sum1;
  pdwExpKey[2] ^= Sum1;
  pdwExpKey[3] ^= Sum1;
  pdwExpKey[4] ^= Sum0;
  pdwExpKey[5] ^= Sum0;
  pdwExpKey[6] ^= Sum0;
  pdwExpKey[7] ^= Sum0;

  CryptonGenEncRoundKey (&EncKey, pdwExpKey);
  CryptonGenDecRoundKey (&DecKey, pdwExpKey);

}

/*********************************************************************/

#define ROTATE(X, n)  ROTL(X, n)


/* masking words for row permutations */
#define    M20    0xcffccffc
#define    M31    0x3ff33ff3
#define    M02    0xfccffccf
#define    M13    0xf33ff33f

/* row-wise bit permutations */
#define RowPerm0(A0)                                  \
    (       (A0 & M20    ) ^ ROTATE(A0 & M31,  8) ^   \
      ROTATE(A0 & M02, 16) ^ ROTATE(A0 & M13, 24) )
#define RowPerm1(A1)                                  \
    (       (A1 & M31    ) ^ ROTATE(A1 & M02,  8) ^   \
      ROTATE(A1 & M13, 16) ^ ROTATE(A1 & M20, 24) )
#define RowPerm2(A2)                                  \
    (       (A2 & M02    ) ^ ROTATE(A2 & M13,  8) ^   \
      ROTATE(A2 & M20, 16) ^ ROTATE(A2 & M31, 24) )
#define RowPerm3(A3)                                  \
    (       (A3 & M13    ) ^ ROTATE(A3 & M20,  8) ^   \
      ROTATE(A3 & M31, 16) ^ ROTATE(A3 & M02, 24) )

/* Masking constants:
MCi = byte-wise 1-bit left rotate of MCi-1 */
#define  MC0  0xacacacac
#define  MC1  0x59595959
#define  MC2  0xb2b2b2b2
#define  MC3  0x65656565

/* Encryption round constants */
#define  ERC00  0xa54ff53a
#define  ERC01  0xe1bee8ac
#define  ERC02  0x1e2ddc1e
#define  ERC03  0x5a9ccf90
#define  ERC04  0x970bc302
#define  ERC05  0xd37ab674
#define  ERC06  0x0fe9a9e6
#define  ERC07  0x4c589d58
#define  ERC08  0x88c790ca
#define  ERC09  0xc536843c
#define  ERC10  0x01a577ae
#define  ERC11  0x3e146b20
#define  ERC12  0x7a835e92

/* byte-wise rotations */
#define BROTL2(A)  ( ((A&0x3f3f3f3f)<<2) ^ ((A&0xc0c0c0c0)>>6) )
#define BROTL4(A)  ( ((A&0x0f0f0f0f)<<4) ^ ((A&0xf0f0f0f0)>>4) )
#define BROTL6(A)  ( ((A&0x03030303)<<6) ^ ((A&0xfcfcfcfc)>>2) )

/* Encrypt roundkey update */
#define EncKeyUpdate0(pdw, A,B,C,D, RC) {  \
    A = ROTATE(A, 24);                     \
    B = ROTATE(B, 16);                     \
    C = BROTL6(C);                         \
    D = BROTL6(D);                         \
    (pdw)[0] = A ^ RC ^ MC0;               \
    (pdw)[1] = B ^ RC ^ MC1;               \
    (pdw)[2] = C ^ RC ^ MC2;               \
    (pdw)[3] = D ^ RC ^ MC3;               \
}

#define EncKeyUpdate1(pdw, E,F,G,H, RC) {  \
    E = BROTL2(E);                         \
    F = BROTL2(F);                         \
    G = ROTATE(G,  8);                     \
    H = ROTATE(H, 16);                     \
    (pdw)[0] = E ^ RC ^ MC0;               \
    (pdw)[1] = F ^ RC ^ MC1;               \
    (pdw)[2] = G ^ RC ^ MC2;               \
    (pdw)[3] = H ^ RC ^ MC3;               \
}


/*################# Generate Encryption Round Keys #################*/

void
CryptonGenEncRoundKey (DWORD * pdwEncKey, DWORD * pdwExpKey)
{
  DWORD A, B, C, D, E, F, G, H;

  A = pdwExpKey[0];
  B = pdwExpKey[1];
  C = pdwExpKey[2];
  D = pdwExpKey[3];
  E = pdwExpKey[4];
  F = pdwExpKey[5];
  G = pdwExpKey[6];
  H = pdwExpKey[7];

  pdwEncKey[0] = A ^ ERC00 ^ MC0;
  pdwEncKey[1] = B ^ ERC00 ^ MC1;
  pdwEncKey[2] = C ^ ERC00 ^ MC2;
  pdwEncKey[3] = D ^ ERC00 ^ MC3;

  pdwEncKey[4] = E ^ ERC01 ^ MC0;
  pdwEncKey[5] = F ^ ERC01 ^ MC1;
  pdwEncKey[6] = G ^ ERC01 ^ MC2;
  pdwEncKey[7] = H ^ ERC01 ^ MC3;

  EncKeyUpdate0 (pdwEncKey + 8, B, C, D, A, ERC02);
  EncKeyUpdate1 (pdwEncKey + 12, H, E, F, G, ERC03);

  EncKeyUpdate0 (pdwEncKey + 16, C, D, A, B, ERC04);
  EncKeyUpdate1 (pdwEncKey + 20, G, H, E, F, ERC05);

  EncKeyUpdate0 (pdwEncKey + 24, D, A, B, C, ERC06);
  EncKeyUpdate1 (pdwEncKey + 28, F, G, H, E, ERC07);

  EncKeyUpdate0 (pdwEncKey + 32, A, B, C, D, ERC08);
  EncKeyUpdate1 (pdwEncKey + 36, E, F, G, H, ERC09);

  EncKeyUpdate0 (pdwEncKey + 40, B, C, D, A, ERC10);
  EncKeyUpdate1 (pdwEncKey + 44, H, E, F, G, ERC11);

  C = ROTATE (C, 24) ^ ERC12 ^ MC0;
  D = ROTATE (D, 16) ^ ERC12 ^ MC1;
  A = BROTL6 (A) ^ ERC12 ^ MC2;
  B = BROTL6 (B) ^ ERC12 ^ MC3;

  pdwEncKey[48] = RowPerm2 (C);
  pdwEncKey[49] = RowPerm3 (D);
  pdwEncKey[50] = RowPerm0 (A);
  pdwEncKey[51] = RowPerm1 (B);
}


/*################# Generate Decryption Round Keys #################*/

/* Decryption round constants */
#define  DRC00  0xa77b876e
#define  DRC01  0x560d4379
#define  DRC02  0xc05bfa1c
#define  DRC03  0x4ffd46bf
#define  DRC04  0xcd97d19e
#define  DRC05  0x984d898d
#define  DRC06  0x4e836004
#define  DRC07  0x31d91c9f
#define  DRC08  0x5edf469a
#define  DRC09  0x06411bc5
#define  DRC10  0xefdfed2c
#define  DRC11  0xb7f9b6e3
#define  DRC12  0x5ca308d2


/* Decrypt roundkey update */
#define DecKeyUpdate0(pdw, A,B,C,D, RC) {  \
    A = BROTL2(A);                         \
    B = ROTATE(B, 16);                     \
    C = ROTATE(C,  8);                     \
    D = BROTL2(D);                         \
    (pdw)[0] = A ^ RC ^ MC0;               \
    (pdw)[1] = B ^ ROTATE(RC, 24) ^ MC1;   \
    (pdw)[2] = C ^ ROTATE(RC, 16) ^ MC2;   \
    (pdw)[3] = D ^ ROTATE(RC,  8) ^ MC3;   \
}

#define DecKeyUpdate1(pdw, E,F,G,H, RC) {  \
    E = BROTL6(E);                         \
    F = ROTATE(F, 16);                     \
    G = ROTATE(G, 24);                     \
    H = BROTL6(H);                         \
    (pdw)[0] = E ^ RC ^ MC0;               \
    (pdw)[1] = F ^ ROTATE(RC, 24) ^ MC1;   \
    (pdw)[2] = G ^ ROTATE(RC, 16) ^ MC2;   \
    (pdw)[3] = H ^ ROTATE(RC,  8) ^ MC3;   \
}


void
CryptonGenDecRoundKey (DWORD * pdwDecKey, DWORD * pdwExpKey)
{
  DWORD A, B, C, D, E, F, G, H;

  C = pdwExpKey[0];
  D = pdwExpKey[1];
  A = pdwExpKey[2];
  B = pdwExpKey[3];

  F = pdwExpKey[4];
  G = pdwExpKey[5];
  H = pdwExpKey[6];
  E = pdwExpKey[7];

  pdwDecKey[48] = C ^ ERC00 ^ MC0;
  pdwDecKey[49] = D ^ ERC00 ^ MC1;
  pdwDecKey[50] = A ^ ERC00 ^ MC2;
  pdwDecKey[51] = B ^ ERC00 ^ MC3;

  A = RowPerm2 (A);
  B = RowPerm1 (B);
  C = RowPerm1 (C);
  D = RowPerm0 (D);

  A = BROTL4 (A);
  B = BROTL2 (B);
  D = BROTL2 (D);

  E = RowPerm0 (E);
  F = RowPerm1 (F);
  G = RowPerm0 (G);
  H = RowPerm2 (H);

  E = BROTL6 (E);
  F = BROTL6 (F);
  G = BROTL4 (G);
  H = BROTL4 (H);

  pdwDecKey[0] = A ^ DRC00 ^ MC0;
  pdwDecKey[1] = B ^ ROTATE (DRC00, 24) ^ MC1;
  pdwDecKey[2] = C ^ ROTATE (DRC00, 16) ^ MC2;
  pdwDecKey[3] = D ^ ROTATE (DRC00, 8) ^ MC3;

  pdwDecKey[4] = E ^ DRC01 ^ MC0;
  pdwDecKey[5] = F ^ ROTATE (DRC01, 24) ^ MC1;
  pdwDecKey[6] = G ^ ROTATE (DRC01, 16) ^ MC2;
  pdwDecKey[7] = H ^ ROTATE (DRC01, 8) ^ MC3;

  DecKeyUpdate0 (pdwDecKey + 8, D, A, B, C, DRC02);
  DecKeyUpdate1 (pdwDecKey + 12, F, G, H, E, DRC03);

  DecKeyUpdate0 (pdwDecKey + 16, C, D, A, B, DRC04);
  DecKeyUpdate1 (pdwDecKey + 20, G, H, E, F, DRC05);

  DecKeyUpdate0 (pdwDecKey + 24, B, C, D, A, DRC06);
  DecKeyUpdate1 (pdwDecKey + 28, H, E, F, G, DRC07);

  DecKeyUpdate0 (pdwDecKey + 32, A, B, C, D, DRC08);
  DecKeyUpdate1 (pdwDecKey + 36, E, F, G, H, DRC09);

  DecKeyUpdate0 (pdwDecKey + 40, D, A, B, C, DRC10);
  DecKeyUpdate1 (pdwDecKey + 44, F, G, H, E, DRC11);
}




#include <string.h>

typedef unsigned long u32;
typedef unsigned char u8;

u32 __stdcall __export
getblocksize ()
{
  return 128;
}

u32 __stdcall __export
getkeysize ()
{
  return 256;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "CRYPTON-v1.0-256-OPT");
}
